/////////////////////////////////////////
// Serial.cpp
//
// Applied Data Systems
// ------------------------------------------------------------------
// This is part of the COMTerm (previously COMTEST) application
// previous numbers: 20-1118x, 700110-0011x
// 
//
// Description
// -----------
// COMTerm is a generic application for testing ADS system COM ports
// data that comes in through a port is echoed to the screen.  Utilities
// are also available for the following functions:
//   - Enable/Disable the RTS line
//   - Enable/Disable the DTR line
//   - Twiddle the RTS line
//   - Twiddle the DTR line
//   - receive/return a file
//   - Send a stream of:
//       - 4 lines of 50 "A" characters
//       - 4 lines of 50 "B" characters
//       - 5 lines of 60 sequential digits
//       - 500 lines of 60 sequential digits
//
// History
// -------
// 20-11182 - May 25, 2000 : akidder
//   - change power-up default to no handshaking
//   - update error messages
//
// 20-11183 - Aug 2, 2002 : ctacke
//   - checks available ports at load
//   - opens lowest available port (not COM1 default any longer)
//
//  
// 20-11184 - Nov 22, 2002 : ctacke
//   - Added a "Relay Port" so data received on one port can be
//     redirected to echo out another port.  The relay port inherits
//     all settings from the primary port.
//
// 720020-11185 COMTEST - Mar 17, 2003 : ctacke
//   - Renumbered
//   - Added 1200/2400 baud support.
//   - Added parity/flow control/stop bits
//   - rearchitected port opening
//   - modified set and clear DTR/CTS
//   - rewrote SendTextDigits to buffer entire output before sending
//   - Added status bar and SendTextDigits timing code
//
// 720020-11186 COMTerm - April 16, 2003 : ctacke
//   - removed data processing from receive thread (no linefeed additions)
//   - added send bulk/byte by byte option
//   - removed send box
//   - removed Send menu item
//   - added Capture
//   - modified receive buffer size and max input amount
//   - renamed to COMTerm
//   - added app icon
/////////////////////////////////////////

#include <windows.h>
#include <commctrl.h>
#include "resource.h"
#include "serial.h"

#define INPUT_BUFFER_SIZE	0x1000
extern TCHAR	szTitle[ ];
extern HWND		hSendBox;
extern HWND		hReceiveBox;
extern HWND		hWnd;

DWORD	speed				= CBR_38400;
int		port_num			= 1;
int		out_port_num		= 0;
INT		flow_cntrl			= ID_FLOWCONTROL_NONE;
DWORD	RtsControl			= RTS_CONTROL_ENABLE;
DWORD	DtrControl			= DTR_CONTROL_ENABLE;
DWORD	FileBufferSize		= 0;
DWORD	FileBufferLength	= 0;
BOOL	FileReceive			= FALSE;
HANDLE	hCommPort			= INVALID_HANDLE_VALUE;
HANDLE	hOutputPort			= INVALID_HANDLE_VALUE;
TCHAR	*FileBuffer;

// communication function prototypes and variables
HANDLE	OpenCommsPort(HWND hWnd, LPWSTR port, DCB Newdcb);
BOOL	CloseCommsPort(HANDLE);
void	ReportCommError(HWND hWnd, LPTSTR lpszMessage);
void	ReportAllocError(HWND hWnd, LPTSTR lpszMessage);
BOOL	SendText(HWND hWnd);
VOID	SendTextFile(HWND hWnd);
VOID	MenuHandler(HWND hWnd, INT idval);
VOID	Twiddle(int DTRorRTS);
BOOL	StartReadThread();
DWORD WINAPI CommReadThreadFunc(LPVOID lpParam);


//////////////////////////////////////////////////////////////////////////
// Communications functions


HANDLE OpenCommsPort(HWND hWnd, LPWSTR port, DCB Newdcb)
{
	HANDLE	hPort		= NULL;
	TCHAR	status[50];

	if(port == NULL)
	{
		return NULL;
	}

	hPort = CreateFile (port, 							// Port Name (Unicode compatible)
							GENERIC_READ | GENERIC_WRITE, 	// Open for Read-Write
            				0,             					// COM port cannot be shared
            				NULL,          					// Always NULL for Windows CE
            				OPEN_EXISTING, 					// For communication resource
            				0,             					// Non-overlapped operation only
            				NULL);        					// Always NULL for Windows CE

	if(hPort == INVALID_HANDLE_VALUE)
	{
		if(GetLastError() != 55)
		{
			ReportCommError(hWnd, _T("opening serial port"));
		}

		_stprintf(status, _T("Could not open %s (error %i)"), port, GetLastError());
		ChangeStatus(status, INFINITE, 0);

		return NULL;
	}

	_stprintf(status, _T("%s successfully opened"), port);
	ChangeStatus(status, 5000, 0);

	DEBUGMSG(TRUE, (_T("\r\nOpened port: %s\r\n"), port));
	// set the timeouts to specify the behavor of reads and writes.
	COMMTIMEOUTS ct;

	ct.ReadIntervalTimeout = MAXDWORD; 
    ct.ReadTotalTimeoutMultiplier = 0; 
    ct.ReadTotalTimeoutConstant = 0; 
    ct.WriteTotalTimeoutMultiplier = 10; 
    ct.WriteTotalTimeoutConstant = 1000; 

	if(!SetCommTimeouts(hPort, &ct))
	{
		ReportCommError(hWnd, TEXT("Setting comm. timeouts."));
		CloseCommsPort(hPort);
		return NULL;
	}

	// get the current communications parameters, and configure baud rate
	DCB dcb;
	dcb.DCBlength = sizeof(DCB);

	if(!GetCommState(hPort, &dcb))
	{
		ReportCommError(hWnd, TEXT("Getting Comms. State."));
		CloseCommsPort(hPort);
		return NULL;
	}

	dcb.BaudRate		= Newdcb.BaudRate;		// set baud rate 
	dcb.fOutxCtsFlow	= Newdcb.fOutxCtsFlow;
	dcb.fOutxDsrFlow    = Newdcb.fOutxDsrFlow;
	dcb.fOutX           = FALSE; 		// no XON/XOFF control
	dcb.fInX            = FALSE;
	dcb.ByteSize        = Newdcb.ByteSize;
	dcb.Parity          = Newdcb.Parity;
	dcb.StopBits        = Newdcb.StopBits;
	dcb.fRtsControl     = Newdcb.fRtsControl;
	dcb.fDtrControl		= Newdcb.fDtrControl;

	if(!SetCommState(hPort, &dcb))
	{
		ReportCommError(hWnd, TEXT("Setting Comms. State."));
		CloseCommsPort(hPort);
		return NULL;
	}

	TCHAR statusString[255];
	_stprintf(statusString, _T("%s  %i  %i,%s,%i"), port, dcb.BaudRate, dcb.ByteSize, dcb.Parity ? _T("Y") : _T("N"), dcb.StopBits == 0 ? 1 : 2 );	
	ChangeStatus(statusString, INFINITE, 1);

	return hPort;
}

BOOL StartReadThread()
{
	// now need to create the thread that will be reading the comms port
	HANDLE hCommReadThread = CreateThread(NULL, 0, CommReadThreadFunc, NULL, 0, NULL);

	if(hCommReadThread == INVALID_HANDLE_VALUE)	
	{
		ReportCommError(hWnd, TEXT("Creating Thread."));
		CloseCommsPort(hCommPort);
		return FALSE;
	}
	else
	{
		SetThreadPriority(hCommReadThread, THREAD_PRIORITY_HIGHEST);

		CloseHandle(hCommReadThread);
	}
	return TRUE;
}

VOID Twiddle(int DTRorRTS)
{
	int i;
	DCB dcb, oldDCB;

	dcb.DCBlength = sizeof(DCB);
	if(!GetCommState(hCommPort, &dcb))
	{
		ReportCommError(hWnd, TEXT("getting serial port state for Twiddle."));
		CloseCommsPort(hCommPort);
		return ;
	}

	oldDCB = dcb;

	// Twiddle the appropriate signal
	for (i=0; i<100; i++) 
	{
		if (DTRorRTS) 
			dcb.fRtsControl     = RTS_CONTROL_ENABLE;
		else
			dcb.fDtrControl     = DTR_CONTROL_ENABLE;

		if(!SetCommState(hCommPort, &dcb))
		{
			ReportCommError(hWnd, TEXT("setting serial port state for Twiddle."));
			CloseCommsPort(hCommPort);
			return ;
		}

		if (DTRorRTS) 
			dcb.fRtsControl     = RTS_CONTROL_DISABLE;
		else
			dcb.fDtrControl     = DTR_CONTROL_DISABLE;

		if(!SetCommState(hCommPort, &dcb))
		{
			ReportCommError(hWnd, TEXT("setting serial port state for Twiddle."));
			CloseCommsPort(hCommPort);
			return ;
		}
	}

	// Return to the original setting
	if(!SetCommState(hCommPort, &oldDCB))
	{
		ReportCommError(hWnd, TEXT("setting serial port state for Twiddle."));
		CloseCommsPort(hCommPort);
		return ;
	}
}

BOOL EnableRTS(BOOL DataFlowOn)
{
	DCB dcb;

	if (DataFlowOn)
		dcb.fRtsControl = RTS_CONTROL_ENABLE;
	else
		dcb.fRtsControl = RTS_CONTROL_DISABLE;

	return SetCommState(hCommPort, &dcb);
}

// Stops the thread that's reading the comms. port, and then closes the
// comms port.

BOOL CloseCommsPort(HANDLE port)
{
	DEBUGMSG(TRUE, (_T("\r\nClosed port\r\n")));

	if(port != INVALID_HANDLE_VALUE)
	{
		CloseHandle(port);
		port = INVALID_HANDLE_VALUE;
		Sleep(1);
	}
	return TRUE;
}

// reports any errors encountered in these communications functions.

void ReportCommError(HWND hWnd, LPTSTR lpszMessage)
{
	TCHAR szBuffer[200];
	wsprintf(szBuffer, TEXT("Communications Error %d \n%s"), 
				GetLastError(),
				lpszMessage);
	ChangeStatus(szBuffer, INFINITE, 0);
}

void ReportAllocError(HWND hWnd, LPTSTR lpszMessage)
{
	TCHAR szBuffer[200];
	wsprintf(szBuffer, TEXT("Alloc Error %d \n%s"), 
				GetLastError(),
				lpszMessage);
	ChangeStatus(szBuffer, INFINITE, 0);
}

// Thread function to read communications port. This thread is responsible
// for reading any incoming text and adding it to the text box.


DWORD WINAPI CommReadThreadFunc(LPVOID lpParam)
{
	TCHAR	cChar[3];
	DWORD	dwBytesRead;
	DWORD	fdwCommMask;
	TCHAR	szBuffer[INPUT_BUFFER_SIZE];
	BYTE	echo;
	int		textLen;

	if(hCommPort == NULL)
	{
		DEBUGMSG(TRUE, (_T("Port NULL in Read Thread.\r\n")));
	}

	SetCommMask (hCommPort, EV_RXCHAR);
	szBuffer[0] = '\0';

	while(hCommPort != INVALID_HANDLE_VALUE)
	{
		WaitCommEvent (hCommPort, &fdwCommMask, 0);
		SetCommMask (hCommPort, EV_RXCHAR);

		// read the next block of data
		do 
		{
			if(!ReadFile(hCommPort,			// where to read from (the open comm port)
					 &echo,					// where to put the character
					 1,					// number of bytes to read
					 &dwBytesRead,			// how many bytes were actually read
					 NULL))					// overlapped I/O not supported
			{
				if (GetLastError() != ERROR_INVALID_HANDLE)
				{
					ReportCommError(NULL, TEXT("reading serial port."));
				}
				return 0;		// terminate thread on first error or port closed
			}
			else if(dwBytesRead > 0)
			{ // add it to the text box
				if(FileReceive) 
				{
					if (FileBufferSize == 0) 
					{
						FileBufferSize = 1000;
						FileBuffer = (TCHAR *) malloc(FileBufferSize * sizeof(TCHAR));

						if (FileBuffer == NULL) 
						{
							ReportAllocError(NULL, TEXT("Out of memory."));
							return 0;
						}
					}

					if (FileBufferSize <= FileBufferLength) 
					{

						FileBufferSize *= 2;
						FileBuffer = (TCHAR *) realloc(FileBuffer, FileBufferSize * sizeof(TCHAR));

						if (FileBuffer == NULL) 
						{
							ReportAllocError(NULL, TEXT("re-allocing space."));
							return 0;
						}

						Sleep(500);  // to make sure flow control works we slow down here!

					} // if (FileBufferSize <= FileBufferLength) 

					FileBuffer[FileBufferLength++] = echo;
				} // if(FileReceive)

				cChar[0] = echo;

				if(echo == 13)
				{
					wcscat(szBuffer, _T("\r\n"));
				}
				else if(echo != 10)
				{
					cChar[1] = '\0';	// null terminate buffer
					wcscat(szBuffer, cChar);
				}
				
				if(captureText)
				{
					fwrite(&cChar, 1, 1, fCaptureFile);
				}

				if(wcslen(szBuffer) >= INPUT_BUFFER_SIZE - (2 * sizeof(TCHAR)))
				{
					// put insertion pointer at end of text box
					SendMessage(hReceiveBox, EM_SETSEL, -1, -1);
					// append text by replacing current selection (and since there is
					// no current selection, text is inserted at current insertion point)
					SendMessage(hReceiveBox, EM_REPLACESEL, FALSE, (LPARAM)szBuffer);
					szBuffer[0] = '\0';
				} 

				// if we're set to relay, send it
				if(out_port_num > 0)
				{
					DWORD dwBytesWritten;

					for(DWORD i = 0; i < dwBytesRead; i++)
					{
						if(!WriteFile(hOutputPort,	// where to write to (the open comm port)
								&echo,				// what to write
								dwBytesRead,		// number of bytes to be written to port
								&dwBytesWritten,	// number of bytes that were actually written
								NULL))				// overlapped I/O not supported			
						{
							DEBUGMSG(TRUE, (_T("Failed to relay out (%i)"), GetLastError()));
							return FALSE;
						}
						if(echo == '\r')
						{
							echo = '\n';
							WriteFile(hOutputPort,	// where to write to (the open comm port)
								&echo,				// what to write
								dwBytesRead,		// number of bytes to be written to port
								&dwBytesWritten,	// number of bytes that were actually written
								NULL);
						}
					} // for(DWORD i = 0; i < dwBytesRead; i++)
				} // if(out_port_num > 0)
			} // else if(dwBytesRead > 0)
		} while(dwBytesRead == 1);

		textLen = GetWindowTextLength(hReceiveBox);
		SendMessage(hReceiveBox, EM_SETSEL, textLen, textLen);
		SendMessage(hReceiveBox, EM_REPLACESEL, FALSE, (LPARAM)szBuffer);

		szBuffer[0] = '\0';
		echo = 0;
	}

	return 0;
}

// sends text in the send text box to the host. Clears out the text box after
// doing so

BOOL SendText(char *sendBuffer)
{
	int i;
	char *p;

	// find out how many characters there are to send
	DWORD dwBytesToWrite = strlen(sendBuffer);
	DWORD dwBytesWritten;

	if(sendFullBuffer)
	{
		// send
		if(!WriteFile(hCommPort, sendBuffer, dwBytesToWrite, &dwBytesWritten, NULL))
		{
			ReportCommError(hWnd, TEXT("sending text."));
			return FALSE;
		}
	}
	else
	{
		p = sendBuffer;

		for(i = 0 ; i < (int)dwBytesToWrite ; i++)
		{
			// send
			if(!WriteFile(hCommPort, p, 1, &dwBytesWritten, NULL))
			{
				ReportCommError(hWnd, TEXT("sending text."));
				return FALSE;
			}
			p++;
		}
	}

	return TRUE;
}

VOID SendTextFile(HWND hWnd)
{
	DWORD dwBytesWritten;

	if(!WriteFile(hCommPort,					// where to write to (the open comm port)
			FileBuffer,							// what to write
			FileBufferLength * sizeof(TCHAR),	// number of bytes to be written to port
			&dwBytesWritten,					// number of bytes that were actually written
			NULL))								// overlapped I/O not supported			
	{
		ReportCommError(hWnd, TEXT("sending text file."));
		return;
	}

	free(FileBuffer);
	FileBufferLength = 0;
	FileBufferSize = 0;
	FileReceive = FALSE;
}


BOOL SendTextN(HWND hWnd,char sendChar,int number)
{
	// find out how many characters there are to send

	DWORD dwBytesToWrite = number;
	DWORD dwBytesWritten, et;
	TCHAR	status[50];
	char	Cr = '\r';

	char *pBuffer, *p;

	pBuffer = (char *)malloc(dwBytesToWrite * sizeof(char));

	if(pBuffer == NULL)
	{
		ReportCommError(hWnd, _T("Insufficient memory for requested operation."));
		return FALSE;
	}

	else
	{
		for(DWORD i = 0; i < dwBytesToWrite; i++)
		{
			pBuffer[i] = sendChar;
		}
		
		et = GetTickCount();

		if(sendFullBuffer)
		{
			// send
			if(!WriteFile(hCommPort, pBuffer, dwBytesToWrite, &dwBytesWritten, NULL))
			{
				ReportCommError(hWnd, TEXT("sending text (chars)."));
				free(pBuffer);
				return FALSE;
			}
		}
		else
		{
			p = pBuffer;

			for(i = 0 ; i < (int)dwBytesToWrite ; i++)
			{
				// send
				if(!WriteFile(hCommPort, p, 1, &dwBytesWritten, NULL))
				{
					ReportCommError(hWnd, TEXT("sending text (chars)."));
					free(pBuffer);
					return FALSE;
				}
				p++;
			}
		}

		free(pBuffer);		
	}

	WriteFile(hCommPort,&Cr, 1, &dwBytesWritten,NULL);  // write a final CR

	et = GetTickCount() - et;

	_stprintf(status, _T("Sent %i bytes in %ims (%d.%d kbps)"), dwBytesToWrite, et, dwBytesToWrite * 8 / et, (dwBytesToWrite * 8 * 10 / et) - ((dwBytesToWrite * 8 / et) * 10));
	ChangeStatus(status, 10000, 0);

	return TRUE;
}

//
// Sends lines of "0123456789" out the active serial port
//
BOOL SendTextDigits(HWND hWnd, int linesToSend)
{
	char	*buffer, *p;
	char	line[7], Digitbuf[]="0123456789";
	int		offset, i, j;
	DWORD	dwBytesWritten, len, et;
	TCHAR	status[50];
		
	len = linesToSend * (60 + 7 + 2);  // 60 chars + line numbers + line feeds

	p = buffer = (char*)LocalAlloc(LPTR, len);

	if(buffer == NULL)
	{
		ReportCommError(hWnd, _T("Insufficient memory for requested operation."));
		return FALSE;
	}

	for(i = 1; i <= linesToSend; i++)
	{
		// Print line number before digits
		sprintf(line, "%04d ",i);		
		offset = strlen(line);
		memcpy(p, line, offset);
		p += offset;
		
		// add the digits
		offset = strlen(Digitbuf);
		for (j=0; j<6; j++)
		{
			memcpy(p, Digitbuf, offset);
			p += offset;
		}
		
		// add a line feed
		sprintf(line, "\r",i);		
		offset = strlen(line);
		memcpy(p, line, offset);
		p += offset;
	}
	
	// null terminate
	p = '\0';

	et = GetTickCount();

	if(sendFullBuffer)
	{
		// send
		if(!WriteFile(hCommPort, buffer, len, &dwBytesWritten, NULL))
		{
			ReportCommError(hWnd, TEXT("sending text (digits)."));
			LocalFree(buffer);
			return FALSE;
		}
	}
	else
	{
		p = buffer;

		for(i = 0 ; i < (int)len ; i++)
		{
			// send
			if(!WriteFile(hCommPort, p, 1, &dwBytesWritten, NULL))
			{
				ReportCommError(hWnd, TEXT("sending text (digits)."));
				LocalFree(buffer);
				return FALSE;
			}
			p++;
		}
	}

	et = GetTickCount() - et;

	_stprintf(status, _T("Sent %i bytes in %ims (%d.%d kbps)"), len, et, len * 8 / et, (len * 8 * 10 / et) - ((len * 8 / et) * 10));
	ChangeStatus(status, 10000, 0);

	LocalFree(buffer);

	return TRUE;
}

BOOL SetEscapeFunc(UINT idEscape)
{
	BOOL bRet=FALSE;
	DCB dcb;

	if (!GetCommState(hCommPort,&dcb))
	{
		return FALSE;
	}

	switch (idEscape)
	{

	case ID_ESC_SET_RTS:
		RETAILMSG(1,(L" Set RTS with EscapeCommFunction() \r\n"));
		dcb.fRtsControl = RTS_CONTROL_ENABLE;
		SetCommState(hCommPort, &dcb);
		break;

	case ID_ESC_CLR_RTS:
		RETAILMSG(1,(L" Clear RTS with EscapeCommFunction() \r\n"));
		dcb.fRtsControl = RTS_CONTROL_DISABLE;
		SetCommState(hCommPort, &dcb);
		break;

	case ID_ESC_SET_DTR:
		bRet = EscapeCommFunction(hCommPort, SETDTR);
		break;

	case ID_ESC_CLR_DTR:
		bRet = EscapeCommFunction(hCommPort, CLRDTR);
		break;
	default:
		break;
	}

	return bRet;

}